""" Base Multivariate Cosine Similarity Entropy function."""
import numpy as np     

def MvCoSiEn(Data, m=None, tau=None, r=0.1, Norm=0, Logx=2):
    """MvCoSiEn  estimates the multivariate cosine similarity entropy of a multivariate dataset.

    .. code-block:: python
    
        MCoSi, Bm = MvCoSiEn(Data) 

    Returns the multivariate cosine similarity entropy estimate (``MCoSi``)
    and the corresponding global probabilities (``Bm``) estimated for the 
    M multivariate sequences in ``Data`` using the default parameters: 
    embedding dimension = 2*np.ones(M,1), time delay = np.ones(M,1), 
    angular threshold = 0.1, logarithm = 2, data normalization = none, 
    
    .. note::

        To maximize the number of points in the embedding process, this algorithm 
        uses N-max(m*tau) delay vectors and **not** N-max(m)*max(tau) as employed 
        in [1][2].
        
  
    .. code-block:: python

         MCoSi, Bm = MvCoSiEn(Data, keyword = value, ...)

    Returns the multivariate cosine similarity entropy estimates (``MCoSi``) 
    estimated from the M multivariate data sequences in ``Data`` using the 
    specified keyword arguments:
        :Data:  - Multivariate dataset, NxM matrix of N (>10) observations (rows) and M (cols) univariate data sequences
        :m:     - Embedding Dimension, a vector of M positive integers
        :tau:   - Time Delay, a vector of M positive integers      
        :r:     - Angular threshold, a value in range [0 < ``r`` < 1]
        :Logx:  - Logarithm base, a positive scalar (enter 0 for natural log) 
        :Norm:  - Normalisation of ``Data``, one of the following integers:
                :0:   no normalisation - default
                :1:   remove median(``Data``) to get zero-median series
                :2:   remove mean(``Data``) to get zero-mean series
                :3:   normalises each sequence in ``Data`` to unit variance and zero mean
                :4:   normalises each sequence in ``Data`` values to range [-1 1]
                
              
    :See also:
        ``CoSiEn``, ``MvDispEn``, ``MvSampEn``, ``MvFuzzEn``, ``MvPermEn``, ``MSEn``
    
    :References:
         [1] H. Xiao, T. Chanwimalueang and D. P. Mandic, 
              "Multivariate Multiscale Cosine Similarity Entropy" 
              IEEE International Conference on Acoustics, Speech and Signal Processing (ICASSP),
              pp. 5997-6001, doi: 10.1109/ICASSP43922.2022.9747282.
    
         [2] Xiao, H.; Chanwimalueang, T.; Mandic, D.P., 
              "Multivariate Multiscale Cosine Similarity Entropy and Its 
              Application to Examine Circularity Properties in Division Algebras."
              Entropy 2022, 24, 1287. 
    
         [3] Ahmed Mosabber Uddin, Danilo P. Mandic
              "Multivariate multiscale entropy: A tool for complexity
              analysis of multichannel data."
              Physical Review E 84.6 (2011): 061918.
    
         [4] Theerasak Chanwimalueang and Danilo Mandic,
              "Cosine similarity entropy: Self-correlation-based complexity
              analysis of dynamical systems."
              Entropy 
              19.12 (2017): 652.
                      
    """
   
    Data = np.squeeze(Data)
    assert Data.shape[0]>10 and Data.ndim==2 and Data.shape[1]>1,  "Data:   must be an NxM numpy matrix where N>10 and M>1"
    N, Dn = Data.shape 
    if m is None:    m = 2*np.ones(Dn, dtype=int)
    if tau is None:  tau = np.ones(Dn, dtype=int)
    m = m.astype(int)
    tau = tau.astype(int) 
  
    assert isinstance(m,np.ndarray) and all(m>0) and m.size==Dn and m.ndim==1, "m:     must be numpy vector of M positive integers"
    assert isinstance(tau,np.ndarray) and all(tau>0) and tau.size==Dn and tau.ndim==1, "tau:   must be numpy vector of M positive integers"
    assert isinstance(r, (int,float)) and r > 0 and r < 1, "r:    must be a value in range 0 < r < 1"
    assert isinstance(Logx,(int,float)) and (Logx>=0), "Logx:     must be a positive value"
    assert isinstance(Norm,int) and np.isin(Norm, np.arange(5)), "Norm:     must be an integer in range [0 4]"
       
    if Logx == 0: Logx = np.exp(1)
    
    if Norm == 1:
        Xi = Data - np.median(Data,axis=0)
    elif Norm == 2:
        Xi = Data - np.mean(Data,axis=0)
    elif Norm == 3:
        Xi = (Data - np.mean(Data,axis=0))/np.std(Data,axis=0)
    elif Norm == 4:
        Xi = (2*(Data - np.min(Data,axis=0))/np.ptp(Data,axis=0)) - 1
    else:
        Xi = Data*1 
       
    Nx = N - max((m-1)*tau)
    Zm = np.zeros((Nx,sum(m)))               
    q = 0 
    for k in range(Dn):
        for p in range(m[k]):
            Zm[:,q] = Xi[p*tau[k]:Nx+p*tau[k],  k]
            q += 1
        
    Num = np.inner(Zm,Zm)
    Mag = np.sqrt(np.diag(Num))
    Den = np.outer(Mag, Mag)

    AngDis = np.arccos(np.triu(Num/Den,1))/np.pi    
    if np.max(np.imag(AngDis)) < 10**-6:
        Bm = np.sum(np.triu(np.round(AngDis,6) < r,1))/(Nx*(Nx-1)/2)    
    else:
        Bm = np.sum(np.triu(np.real(AngDis) < r,1))/(Nx*(Nx-1)/2)
        print('Warning: Complex values ignored')
    if Bm == 1 or Bm == 0:
        CoSi = np.NaN
    else:
        CoSi = -(Bm*np.log(Bm)/np.log(Logx)) - ((1-Bm)*np.log(1-Bm)/np.log(Logx))

    return CoSi, Bm

    
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Bidimensional Espinosa Entropy function."""
import numpy as np    

def EspEn2D(Mat, m=None, tau=1, r=20, ps=0.7, Logx=np.exp(1), Lock=True):
    """EspEn2D  Estimates the bidimensional Espinosa entropy of a data matrix.

    .. code-block:: python
    
        Esp2D = EspEn2D(Mat) 
        
    Returns the bidimensional Espinosa entropy estimate (``Esp2D``) 
    estimated for the data matrix (``Mat``) using the default parameters: 
    time delay = 1, tolerance threshold = 20, percentage similarity = 0.7
    logarithm = natural, matrix template size = [floor(H/10) floor(W/10)]  
    (where H and W represent the height (rows) and width (columns) of 
    the data matrix ``Mat``) 
    ** The minimum number of rows and columns of ``Mat`` must be > 10.
 
     .. code-block:: python
     
         Esp2D = EspEn2D(Mat, keyword = value, ...)
    
    Returns the bidimensional Espinosa entropy (``Esp2D``) estimates for the data
    matrix (``Mat``) using the specified 'keyword' arguments:
        :m:     - Template submatrix dimensions, an integer scalar (for sub-matrix with same height and width) or a two-element tuple of integers (height, width) with a minimum value > 1. (default: [floor(H/10) floor(W/10)])
        :tau:   - Time Delay, a positive integer > 1   (default: 1)
        :r:     - Tolerance Threshold, a positive scalar   (default: 20)
        :ps:    - Percentage similarity, a scalar in range [0 1] (default: 0.7)        
        :Logx:  - Logarithm base, a positive scalar    (default: natural)        
        :Lock:  - By default, ``EspEn2D`` only permits matrices with a maximum  size of 128 x 128 to prevent RAM overload. 
              e.g. For ``Mat`` = [200 x 200], ``m`` = 3, and ``tau`` = 1, ``EspEn2D`` 
              creates a vector of 753049836 elements. To enable matrices greater than [128 x 128] elements, set ``Lock = False`` (default: True)
              ``CAUTION: unlocking the permitted matrix size may cause memory``
              ``errors that could lead your Python IDE to crash.``
    
    :See also:
        ``SampEn2D``, ``FuzzEn2D``, ``DistEn2D``, ``DispEn2D``
    
    :References:
        [1] Ricardo Espinosa, et al.,
          "Two-dimensional EspEn: A New Approach to Analyze Image Texture 
          by Irregularity." 
          Entropy,
          23:1261 (2021)
                                 
    """
            
    Mat = np.squeeze(Mat)
    NL,NW = Mat.shape    
    if m is None:
        m = np.array(Mat.shape)//10
           
    assert Mat.ndim==2 and min(Mat.shape)>10 , \
    "Mat:   must be a 2D numpy array with height & width > 10"
    assert (isinstance(m,int) and m>1) or (isinstance(m, (np.ndarray,tuple))  
                                           and len(m)==2 and min(m)>1), \
    "m:     must be an integer > 1, or a 2 element tuple of integers > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(r,(int,float)) and r>0, "r:     must be a positive value"
    assert isinstance(ps,float) and ps>=0 and ps<=1, "r:     must be a positive value"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Lock,bool) and ((Lock==True and max(Mat.shape)<=128) or Lock==False), \
    "Lock:      To prevent memory storage errors, matrix width & length must \
    have <= 128 elements. To estimate EspEn2D for the current matrix (%d x %d) \
    change Lock to False. \
    CAUTION: unlocking the permitted matrix size may cause memory \
    errors that could lead your Python IDE to crash.."%(NW,NL)
    
    if isinstance(m,int):
        mL = int(m); mW = int(m)        
    else:
        mL = int(m[0]); mW = int(m[1])       
        
    NL = NL - (mL-1)*tau
    NW = NW - (mW-1)*tau
    X = np.zeros((NL*NW,mL,mW))
    p = 0
    for k in range(NL):        
        for n in range(NW):
            X[p,:,:] = Mat[k:(mL*tau)+k:tau,n:(mW*tau)+n:tau]
            p += 1              
            
    if p != NL*NW:
        print('Warning: Potential error with submatrix division.')        
    Ny = int(p*(p-1)/2)
    if Ny > 300000000:
        print('Warning: Number of pairwise distance calculations is ' + str(Ny))
    
    Cij = -np.ones((p-1,p-1))
    for k in range(p-1):
        Temp = abs(X[k+1:,:,:] - X[k,:,:]) <= r
        Cij[:p-(k+1),k] = np.sum(Temp,axis=(1,2))
     
    Dm = np.sum((Cij/(mL*mW))>=ps)/(p*(p-1)/2)
    Esp2D = -np.log(Dm)/np.log(Logx)
    return Esp2D

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"Base composite cross-Multiscale Entropy function."
import numpy as np  
from matplotlib.pyplot import figure, axes, show
from copy import deepcopy

def cXMSEn(Sig1, Sig2, Mbjx,  Scales=3, RadNew=0, Refined=False, Plotx=False):
    """cXMSEn  returns the composite (or refined-composite) multiscale cross-entropy between two univariate data sequences.
 
    .. code-block:: python
        
        MSx, CI = cXMSEn(Sig1, Sig2, Mobj) 
        
    Returns a vector of composite multiscale cross-entropy values (``MSx``) 
    between two univariate data sequences contained in ``Sig1`` and ``Sig2`` using the 
    parameters specified by the multiscale object (``Mobj``) using the composite 
    multiscale method (cMSE) over 3 temporal scales.
    
    .. code-block:: python 
        
        MSx, CI = cXMSEn(Sig1, Sig2, Mobj, Refined = True) 
        
    Returns a vector of refined-composite multiscale cross-entropy values (``MSx``) for the data 
    sequences (``Sig1``, ``Sig2``) using the parameters specified by the multiscale object 
    (``Mobj``) using the refined-composite multiscale entropy method (rcMSE) over 3 temporal
    scales. When ``Refined == True``, the base entropy method must be ``XSampEn`` or ``XFuzzEn``.
    If the entropy method is ``XSampEn``, cXMSEn employs the method described in [7]. 
    If the entropy method is ``XFuzzEn``, cXMSEn employs the method described in [8].    
     
    .. code-block:: python
        
        MSx, CI = cXMSEn(Sig1, Sig2, Mobj, keyword = value, ...)
        
    Returns a vector of composite multiscale cross-entropy values (``MSx``) 
    between the data sequences contained in ``Sig1`` and ``Sig2`` using the parameters
    specified by the multiscale object (Mobj) and the following 'keyword' arguments:
        :Scales:   - Number of temporal scales, an integer > 1   (default: 3)
        :RadNew:   - Radius rescaling method, an integer in the range [1 4].
                     When the cross-entropy specified by ``Mobj`` is ``XSampEn`` or ``XApEn``, RadNew rescales the radius threshold in each sub-sequence
                     at each time scale (Ykj). If a radius value (``r``) is specified by ``Mobj``, this becomes the rescaling coefficient, otherwise
                     it is set to 0.2 (default). The value of RadNew specifies one of the following methods:
                         
                        * [1] Pooled Standard Deviation   - ``r*std(Ykj)``
                        * [2] Pooled Variance             - ``r*var(Ykj)``
                        * [3] Mean Absolute Deviation     - ``r*mad(Ykj)``
                        * [4] Median Absolute Deviation   - ``r*mad(Ykj,1)``
                     
        :Refined:  - Refined-composite XMSEn method. When ``Refined == True`` and the cross-entropy function specified by
                    ``Mobj`` is ``XSampEn`` or ``XFuzzEn``,  ``cXMSEn`` returns the refined-composite multiscale entropy (rcXMSEn) [default: False]
        :Plotx:    - When ``Plotx == True``, returns a plot of the entropy value at each time scale (i.e. the multiscale entropy curve) [default: False]
    
    :See also:
        ``MSobject``, ``XMSEn``, ``rXMSEn``, ``hXMSEn``, ``XSampEn``, ``XApEn``, ``MSEn``, ``cMSEn``, ``rMSEn``
    
    :References:
        [1] Rui Yan, Zhuo Yang, and Tao Zhang,
            "Multiscale cross entropy: a novel algorithm for analyzing two
            time series." 
            5th International Conference on Natural Computation. 
            Vol. 1, pp: 411-413 IEEE, 2009.
    
        [2] Yi Yin, Pengjian Shang, and Guochen Feng, 
            "Modified multiscale cross-sample entropy for complex time 
            series."
            Applied Mathematics and Computation 
            289 (2016): 98-110.
    
        [3] Madalena Costa, Ary Goldberger, and C-K. Peng,
            "Multiscale entropy analysis of complex physiologic time series."
            Physical review letters
            89.6 (2002): 068102.
    
        [4] Antoine Jamin, et al,
            "A novel multiscale cross-entropy method applied to navigation 
            data acquired with a bike simulator." 
            41st annual international conference of the IEEE EMBC
            IEEE, 2019.
    
        [5] Antoine Jamin and Anne Humeau-Heurtier. 
            "(Multiscale) Cross-Entropy Methods: A Review." 
            Entropy 
            22.1 (2020): 45.
    
        [6] Shuen-De Wu, et al.,
            "Time series analysis using composite multiscale entropy." 
            Entropy 
            15.3 (2013): 1069-1084.
            
        [7] Shuen-De Wu, et al.,
            "Analysis of complex time series using refined composite 
            multiscale entropy." 
            Physics Letters A 
            378.20 (2014): 1369-1374.
            
        [8] Hamed Azami et al.,
            "Refined multiscale fuzzy entropy based on standard deviation 
            for biomedical signal analysis"
            Med Biol Eng Comput 
            55 (2017):2037–2052
    
    """
    
    Mobj = deepcopy(Mbjx)    
    S1 = np.squeeze(Sig1)
    S2 = np.squeeze(Sig2)
    assert Sig1.ndim==1 and Sig2.ndim==1,   "Sig1/Sig2:   must be numpy vectors (N>10)" 
    N1 = S1.shape[0]
    N2 = S2.shape[0]

    assert N1>10 and N2>10,  "Sig1/Sig2:   Each sequence must be a numpy vector (N>10)"     
    assert isinstance(Mobj,object) and Mobj.Func.__name__[0]=='X', "Mobj:  must \
    be a multiscale entropy object created with the function EntropyHub.MSobject \
        and the base estimator must be a CROSS-entropy method."    
    assert isinstance(Scales,int) and Scales>1, "Scales:    must be an integer > 1"
    assert isinstance(Refined, bool) and ((Refined==True and Mobj.Func.__name__ 
                    in ['XSampEn','XFuzzEn'])  or Refined==False), \
    "Refined:       must be a True or False. If Refined==True, Mobj.Func must be XSampEn or XFuzzEn"
    assert (np.isin(RadNew,np.arange(1,5)) and Mobj.Func.__name__ in \
    ['XSampEn','XApEn']) or RadNew==0, "RadNew:     must be an integer in range \
    [1 4] and entropy function must be 'XSampEn' or 'XApEn'"
    assert isinstance(Plotx, bool), "Plotx:    must be boolean - True or False"
        
    MSx = np.zeros(Scales)       
    if RadNew > 0:
        if RadNew == 1:
            Rnew = lambda x,y: np.sqrt((np.var(x)*(len(x)-1) + np.var(y)*(len(y)-1))/(len(x)+len(y)-1))
        elif RadNew == 2:
            Rnew = lambda x,y: (np.var(x)*(len(x)-1) + np.var(y)*(len(y)-1))/(len(x)+len(y)-1)
        elif RadNew == 3:
            Rnew = lambda x,y: np.mean(abs(np.hstack((x,y))-np.mean(np.hstack((x,y)))))
        elif RadNew == 4:
            Rnew = lambda x,y: np.median(abs(np.hstack((x,y))-np.median(np.hstack((x,y)))))   
                
        try:
            Cx = Mobj.Kwargs.get('r')*1
        except:
            Cy = ['Pooled Standard Deviation','Pooled Variance','Joint Mean Abs Deviation', 'Joint Median Abs Deviation']
            print('WARNING: No radius value provided.\nDefault set to ' \
                  '0.2*(%s) of each new time-series.'%Cy[RadNew-1])              
            Cx = .2
                        
    if Mobj.Func.__name__ in 'XSampEn': Mobj.Kwargs['Vcp'] = False 

    if Refined and Mobj.Func.__name__ == 'XFuzzEn':
        Tx = 1
        if 'Logx' in Mobj.Kwargs.keys(): Logx = Mobj.Kwargs['Logx']
        else: Logx = np.e
        
    elif Refined and Mobj.Func.__name__ == 'XSampEn':
        Tx = 0
        if 'Logx' in Mobj.Kwargs.keys(): Logx = Mobj.Kwargs['Logx']
        else: Logx = np.e
    
    else:  Tx = 0
       
    for T in range(1,Scales+1):
        Temp = modified(S1, S2, T, Tx)
        
        N1 = T*(len(Temp[0])//T)
        N2 = T*(len(Temp[1])//T)
        
        Temp3 = np.zeros(T)
        Temp2 = np.zeros(T)
        
        for k in range(T):
            print(' .', end='')
            if RadNew > 0:
                Mobj.Kwargs.update({'r': Cx*Rnew(Temp[0][k:N1:T], Temp[1][k:N2:T])})    
                
            if Refined:
                _, Ma, Mb = Mobj.Func(Temp[0][k:N1:T],Temp[1][k:N2:T],**Mobj.Kwargs)
                Temp2[k] = Ma[-1]
                Temp3[k] = Mb[-1]
            else:
                Temp2 = Mobj.Func(Temp[0][k:N1:T],Temp[1][k:N2:T],**Mobj.Kwargs)
                
                if isinstance(Temp2,tuple):
                    if isinstance(Temp2[0],(int,float)):
                        Temp3[k] = Temp2[0]
                    else:
                        Temp3[k] = Temp2[0][-1]       
                elif isinstance(Temp2,(int,float)):
                    Temp3[k] = Temp2
                elif isinstance(Temp2,np.ndarray):
                    Temp3[k] = Temp2[-1]                     
                            
        if Refined and Tx==0:
            MSx[T-1] = -np.log(sum(Temp2)/sum(Temp3))/np.log(Logx)
        elif Refined and Tx==1:
            MSx[T-1] = -np.log(sum(Temp3)/sum(Temp2))/np.log(Logx)
        else:
            MSx[T-1] = np.mean(Temp3)
    
    CI = sum(MSx)
    if np.any(np.isnan(MSx)):
        print('Some entropy values may be undefined.')
    
    print('\n')
            
    if Plotx == 1:
        if Refined:
            strx = 'Refined-Composite Cross-'
        else:
            strx = 'Composite Cross-'
        
        figure()
        ax1 = axes()   
        ax1.plot(np.arange(1,Scales+1), MSx, color=(8/255, 63/255, 77/255), linewidth=3)
        ax1.scatter(np.arange(1,Scales+1), MSx, 60, color=(1,0,1))
        ax1.set_xlabel('Scale Factor',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
        ax1.set_ylabel('Entropy Value',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
        ax1.set_title('%sMultiscale %s'%(strx,Mobj.Func.__name__), 
                     fontsize=16,fontweight='bold',color=(7/255, 54/255, 66/255))       
        show()

    return MSx, CI

def modified(Za,Zb,sx,Tx):    
    if Tx==1:
        T1 = np.asarray([np.std(Za[x:x+sx]) for x in range(len(Za)-sx+1)])
        T2 = np.asarray([np.std(Zb[x:x+sx]) for x in range(len(Zb)-sx+1)])

    else: 
        T1 = np.convolve(Za,np.ones(sx),'valid')/sx
        T2 = np.convolve(Zb,np.ones(sx),'valid')/sx
    
    return T1, T2


"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


""" Base Multivariate Fuzzy Entropy function."""
import numpy as np       

def MvFuzzEn(Data, m=None, tau=None, r=(0.2,2.0), Fx='default', Norm=False, Logx=np.exp(1)):
    """MvFuzzEn  estimates the multivariate fuzzy entropy of a multivariate dataset.

    .. code-block:: python
    
        MFuzz, B0, Bt, B1 = MvFuzzEn(Data) 

    Returns the multivariate fuzzy entropy estimate (``MFuzz``) and the 
    average vector distances (``m``: ``B0``; joint total 
    ``m+1`` subspace: ``Bt``; all possible ``m+1`` subspaces: ``B1``),
    from the M multivariate sequences in ``Data`` using the default parameters: 
    embedding dimension = 2*np.ones(M,1), time delay = np.ones(M,1), 
    fuzzy membership function = "default", fuzzy function parameters= [0.2, 2],
    logarithm = natural, data normalization = false,     
    
    .. attention::

        The entropy value returned as ``MFuzz`` is estimated using the "full" 
        method [i.e.  -log(Bt/B0)] which compares delay vectors across all possible ``m+1`` 
        expansions of the embedding space as applied in [1][3]. Contrary to
        conventional definitions of sample entropy, this method does not provide a
        lower bound of 0!!
        Thus, it is possible to obtain negative entropy values for multivariate 
        fuzzy entropy, even for stochastic processes...
        
        Alternatively, one can calculate ``MFuzz`` via the "naive" method, 
        which ensures a lower bound of 0, by using the average vector distances
        for an individual ``m+1`` subspace (B1) [e.g. -log(B1(1)/B0)],
        or the average for all ``m+1`` subspaces [i.e. -log(mean(B1)/B0)].

    .. note::
        
        To maximize the number of points in the embedding process, this algorithm 
        uses N-max(m*tau) delay vectors and **not** N-max(m)*max(tau) as employed 
        in [1] and [3].
    
    
    .. code-block:: python
    
        MFuzz, B0, Bt, B1 = MvFuzzEn(Data, name, value, ...)

    Returns the multivariate fuzzy entropy estimates (``MFuzz``) estimated 
    from the M multivariate data sequences in ``Data`` using the specified
    keyword arguments:
        :Data:  - Multivariate dataset, NxM matrix of N (>10) observations (rows) and M (cols) univariate data sequences 
        :m:     - Embedding Dimension, a vector of M positive integers
        :tau:   - Time Delay, a vector of M positive integers
        :Fx:    - Fuzzy function name, one of the following strings:
              {``'sigmoid'``, ``'modsampen'``, ``'default'``, ``'gudermannian'``, ``'bell'``, ``'triangular'``, ``'trapezoidal1'``, ``'trapezoidal2'``, ``'z_shaped'``, ``'gaussian'``, ``'constgaussian'``}
        :r:     - Fuzzy function parameters, a 1 element scalar or a 2 element vector of positive values. The ``r`` parameters for each fuzzy  function are defined as follows  (default: [.2 2])
                      :sigmoid:     
                                  * r(1) = divisor of the exponential argument
                                  * r(2) = value subtracted from argument (pre-division)
                      :modsampen:    
                                  *  r(1) = divisor of the exponential argument
                                  *  r(2) = value subtracted from argument (pre-division)
                      :default:      
                                  *  r(1) = divisor of the exponential argument
                                  *  r(2) = argument exponent (pre-division)
                      :gudermannian: 
                                  *  r  = a scalar whose value is the numerator of argument to gudermannian function - GD(x) = atan(tanh(r/x)). 
                                     GD(x) is normalised to have a maximum value of 1.                       
                      :triangular:  
                                  * r = a positive scalar whose value is the threshold 
                                    (corner point) of the triangular function.
                      :trapezoidal1:  
                                  * r = a positive scalar whose value corresponds
                                    to the upper (2r) and lower (r) corner points of the trapezoid.
                      :trapezoidal2:  
                                  * r(1) = a value corresponding to the upper corner point of the trapezoid.
                                  * r(2) = a value corresponding to the lower corner point of the trapezoid.
                      :z_shaped:  
                                  * r = a scalar whose value corresponds to the
                                    upper (2r) and lower (r) corner points of the z-shape.
                      :bell:  
                                  * r(1) = divisor of the distance value
                                  * r(2) = exponent of generalized bell-shaped function
                      :gaussian:  
                                  * r = a scalar whose value scales the slope of the Gaussian curve.
                      :constgaussian:  
                                  * r = a scalar whose value defines the lower 
                                    threshod and shape of the Gaussian curve.
        :Norm:  - Normalisation of all M sequences to unit variance, a boolean
        :Logx:  - Logarithm base, a positive scalar  [default: natural]
      
    For further information on the keyword arguments, see the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_.
    
    :See also:
        ``MvSampEn``, ``FuzzEn``, ``XFuzzEn``, ``FuzzEn2D``, ``MSEn``, ``MvPermEn``.
    
    :References:
       [1] Ahmed, Mosabber U., et al. 
            "A multivariate multiscale fuzzy entropy algorithm with application
            to uterine EMG complexity analysis." 
            Entropy 19.1 (2016): 2.
    
       [2] Azami, Alberto Fernández, Javier Escudero. 
            "Refined multiscale fuzzy entropy based on standard deviation for 
            biomedical signal analysis." 
            Medical & biological engineering & computing 55 (2017): 2037-2052.
    
       [3] Ahmed Mosabber Uddin, Danilo P. Mandic
            "Multivariate multiscale entropy analysis."
            IEEE signal processing letters 19.2 (2011): 91-94.
    """
   
    Data = np.squeeze(Data)
    assert Data.shape[0]>10 and Data.ndim==2 and Data.shape[1]>1,  "Data:   must be an NxM numpy matrix where N>10 and M>1"
    N, Dn = Data.shape 
    if m is None:    m = 2*np.ones(Dn, dtype=int)
    if tau is None:  tau = np.ones(Dn, dtype=int)
    m = m.astype(int)
    tau = tau.astype(int) 
  
    assert isinstance(m,np.ndarray) and all(m>0) and m.size==Dn and m.ndim==1, "m:     must be numpy vector of M positive integers"
    assert isinstance(tau,np.ndarray) and all(tau>0) and tau.size==Dn and tau.ndim==1, "tau:   must be numpy vector of M positive integers"
    assert isinstance(r,(int,float)) or ((r[0] >= 0) and len(r) ==2), "r:     must be a scalar or 2 element tuple of positive values"
    assert Fx.lower() in ['default','sigmoid','modsampen','gudermannian',
                          'bell', 'gaussian', 'constgaussian', 'triangular' ,
                          'trapezoidal1', 'trapezoidal2', 'z_shaped'] \
            and isinstance(Fx,str), """Fx:    must be one of the following strings -
            'default', 'sigmoid', 'modsampen', 'gudermannian', 'bell', 'z_shaped',
            'triangular', 'trapezoidal1','trapezoidal2','gaussian','constgaussian'"""
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be a Boolean"    

    if Norm: Data = Data/np.std(Data,axis=0)
    
    Fun = globals()[Fx.lower()]       
    Nx = N - max((m-1)*tau)
    Ny = N - max(m*tau)
    Vex = np.zeros((Nx,sum(m)))
    q = 0
    for k in range(Dn):
        for p in range(m[k]):
            Vex[:,q] = Data[p*tau[k]:Nx+p*tau[k],  k]
            q += 1
            
    Count0 = Distx(Vex - np.mean(Vex,axis=1,keepdims=True), r, Fun)
    B0 = np.sum(Count0)/(Nx*(Nx-1)/2)
            
    B1 = np.zeros(Dn)
    Vez = np.inf*np.ones((1,sum(m)+1));
    Temp = np.cumsum(m)
    for k in range(Dn):
        Sig = np.expand_dims(Data[m[k]*tau[k]:Ny+m[k]*tau[k], k],1)
        Vey = np.hstack((Vex[:Ny, :Temp[k]], Sig, Vex[:Ny, Temp[k]:]))
        Vez = np.vstack((Vez, Vey))
        Count1 = Distx(Vey - np.mean(Vey,axis=1,keepdims=True), r, Fun)
        B1[k] = np.sum(Count1)/(Ny*(Ny-1)/2)
    Vez = Vez[1:,:]
    Count1 = Distx(Vez - np.mean(Vez,axis=1,keepdims=True), r, Fun)
    Bt = np.sum(Count1)/(Dn*Ny*((Dn*Ny)-1)/2)
       
    with np.errstate(divide='ignore', invalid='ignore'):
        MFuzz = -np.log(Bt/B0)/np.log(Logx) 
   
    return MFuzz, B0, Bt, B1    


def Distx(Vex, r, Fun):
    Nt = Vex.shape[0]
    Counter = np.zeros((Nt-1,Nt-1))
    for x in range(Nt-1):
        Counter[x,x:] = Fun(np.max(abs(Vex[x+1:,:] - Vex[x,:]), axis=1), r)
    return Counter


def sigmoid(x,r):
    assert isinstance(r,tuple), 'When Fx = "Sigmoid", r must be a two-element tuple.'
    y = 1/(1 + np.exp((x-r[1])/r[0]))
    return y  
def default(x,r):   
    assert isinstance(r,tuple), 'When Fx = "Default", r must be a two-element tuple.'
    y = np.exp(-(x**r[1])/r[0])
    return y     
def modsampen(x,r):
    assert isinstance(r,tuple), 'When Fx = "Modsampen", r must be a two-element tuple.'
    y = 1/(1 + np.exp((x-r[1])/r[0]));
    return y    
def gudermannian(x,r):
    if r <= 0:
        raise Exception('When Fx = "Gudermannian", r must be a scalar > 0.')
    y = np.arctan(np.tanh(r/x))    
    y = y/np.max(y)    
    return y    
def triangular(x, r):
    assert np.size(r)==1, 'When Fx = "Triangular", r must be a scalar > 0.'
    y = 1 - (x/r)
    y[x > r] = 0
    return y
def trapezoidal1(x, r):
    assert np.size(r)==1, 'When Fx = "Trapezoidal1", r must be a scalar > 0.'
    y = np.zeros(np.shape(x))
    y[x <= r*2] = 2 - (x[x <= r*2]/r)
    y[x <= r] = 1
    return y
def trapezoidal2(x, r):
    assert isinstance(r,tuple), 'When Fx = "Trapezoidal2", r must be a two-element tuple.'
    y = np.zeros(np.shape(x))
    y[x <= max(r)] = 1 - (x[x <= max(r)] - min(r))/(max(r)-min(r))
    y[x <= min(r)] = 1
    return y
def z_shaped(x, r):
    assert np.size(r)==1, 'When Fx = "Z_shaped", r must be a scalar > 0.'
    y = np.zeros(np.shape(x))
    y[x <= 2*r] = 2*(((x[x <= 2*r] - 2*r)/r)**2)
    y[x <= 1.5*r] = 1 - (2*(((x[x <= 1.5*r] - r)/r)**2))
    y[x <= r] = 1
    return y
def bell(x, r):
    assert isinstance(r,tuple), 'When Fx = "Bell", r must be a two-element tuple.'
    y = 1/(1 + abs(x/r[0])**(2*r[1]))
    return y
def gaussian(x, r):
    assert np.size(r)==1, 'When Fx = "Gaussian", r must be a scalar > 0.'
    y = np.exp(-((x**2)/(2*(r**2))))
    return y
def constgaussian(x, r):
    assert np.size(r)==1, 'When Fx = "ConstGaussian", r must be a scalar > 0.'
    y = np.ones(np.shape(x))
    y[x > r] = np.exp(-np.log(2)*((x[x > r] - r)/r)**2)
    return y


"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


""" Base Distribution Entropy function."""
import numpy as np
from scipy.stats import skew

def DistEn(Sig,  m=2, tau=1, Bins='Sturges', Logx=2, Norm=True):
    """DistEn  estimates the distribution entropy of a univariate data sequence.

    .. code-block:: python
        
        Dist, Ppi = DistEn(Sig) 
        
    Returns the distribution entropy estimate (``Dist``) and the
    corresponding distribution probabilities (``Ppi``) estimated from the data 
    sequence (``Sig``)  using the default  parameters: 
    embedding dimension = 2, time delay = 1, binning method = 'Sturges',
    logarithm = base 2, normalisation = w.r.t # of histogram bins
 
    .. code-block:: python
    
        Dist, Ppi = DistEn(Sig, keyword = value, ...)
        
    Returns the distribution entropy estimate (``Dist``) estimated from the data
    sequence (``Sig``) using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, a positive integer
        :tau:   - Time Delay, a positive integer
        :Bins:  - Histogram bin selection method for distance distribution, one of the following:
            
                  * an integer > 1 indicating the number of bins, 
                  * or one of the following strings {``'sturges'``, ``'sqrt'``, ``'rice'``, ``'doanes'``} [default: ``'sturges'``]
              
        :Logx:  - Logarithm base, a positive scalar (enter 0 for natural log) 
        :Norm:  - Normalisation of ``Dist`` value, a boolean:
            * [False]  no normalisation.
            * [True]   normalises w.r.t # of histogram bins - default
 
    :See also:
        ``XDistEn``, ``DistEn2D``, ``MSEn``, ``K2En``
  
    :References:
        [1] Li, Peng, et al.,
            "Assessing the complexity of short-term heartbeat interval 
            series by distribution entropy." 
            Medical & biological engineering & computing 
            53.1 (2015): 77-87. 
    
    """
    
    Sig = np.squeeze(Sig)
    if Logx == 0:
         Logx = np.exp(1)
    
    assert Sig.shape[0]>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 0), "m:     must be an integer > 0"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(Logx,(int,float)) and Logx > 0, "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - True or False"    
    assert (isinstance(Bins,int) and Bins>1) \
            or (Bins.lower() in ['sturges','sqrt','rice','doanes']), \
            "Bins:    an integer > 1, or can be one of the following strings - \
            'sturges', 'sqrt', 'rice', 'doanes'"    
    
    Nx = Sig.shape[0] - ((m-1)*tau)   
    Zm = np.zeros((Nx,m))
    for n in np.arange(m):
        Zm[:,n] = Sig[n*tau:Nx+(n*tau)]
        
    DistMat = np.zeros(round(Nx*(Nx-1)/2))
    for k in range(1,Nx):
        Ix = (int((k-1)*(Nx - k/2)), int(k*(Nx-((k+1)/2))))
        DistMat[Ix[0]:Ix[1]] = np.max(abs(np.tile(Zm[k-1,:],(Nx-k,1)) - Zm[k:,:]),axis=1)
        
    Ny = len(DistMat)
    if isinstance(Bins, str):
        if Bins.lower() == 'sturges':
            Bx = np.ceil(np.log2(Ny) + 1)
        elif Bins.lower() == 'rice':
            Bx = np.ceil(2*(Ny**(1/3)))
        elif Bins.lower() == 'sqrt':
            Bx = np.ceil(np.sqrt(Ny))
        elif Bins.lower()== 'doanes':
            sigma = np.sqrt(6*(Ny-2)/((Ny+1)*(Ny+3)))
            Bx = np.ceil(1+np.log2(Ny)+np.log2(1+abs(skew(DistMat)/sigma)))
        else:
            raise Exception('Please enter a valid binning method')               
    else:
        Bx = Bins
        
    Ppi,_ = np.histogram(DistMat,int(Bx))        
    Ppi = Ppi/Ny    
    if round(sum(Ppi),6) != 1:
        print('Warning: Potential error estimating probabilities.')
        Ppi = Ppi[Ppi!=0]
    elif any(Ppi==0):
        print('Note: '+str(sum(Ppi==0))+'/'+str(len(Ppi))+' bins were empty')
        Ppi = Ppi[Ppi!=0]
           
    Dist = -sum(Ppi*np.log(Ppi)/np.log(Logx))
    if Norm:
        Dist = Dist/(np.log(Bx)/np.log(Logx))
      
    return Dist, Ppi
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


""" Base Permutation Entropy function."""
import numpy as np
from itertools import permutations as prm 
from scipy.signal import hilbert


def PermEn(Sig, m=2, tau=1, Logx=2, Norm=False, Typex='none', tpx=-1):
    """PermEn  estimates the permutation entropy of a univariate data sequence.

    .. code-block:: python
    
        Perm, Pnorm, cPE = PermEn(Sig) 
        
    Returns the permuation entropy estimates (``Perm``), the normalised
    permutation entropy (``Pnorm``) and the conditional permutation entropy (``cPE``)
    for ``m`` = [1,2] estimated from the data sequence (``Sig``) using the default 
    parameters: embedding dimension = 2, time delay = 1, 
    logarithm = base 2, normalisation = w.r.t #symbols (``m-1``)
    Note: using the standard PermEn estimation, ``Perm`` =0 when ``m`` =1. 
    It is recommeneded that signal length, N > 5m! 
    (see [8] and Amigo et al., Europhys. Lett. 83:60005, 2008)
    
    .. code-block:: python
    
        Perm, Pnorm, cPE = PermEn(Sig, m)
        
    Returns the permutation entropy estimates (``Perm``) estimated from the data
    sequence (``Sig``) using the specified embedding dimensions = [1,..., ``m``] 
    with other default parameters as listed above.
 
    .. code-block:: python
    
        Perm, Pnorm, cPE = PermEn(Sig, keyword = value, ...)
        
    Returns the permutation entropy estimates (``Perm``) for dimensions = [1,..., ``m``]
    estimated from the data sequence (``Sig``) using the specified 'keyword' arguments:        
        :m:     - Embedding Dimension, an integer > 1
        :tau:   - Time Delay, a positive integer
        :Logx:  - Logarithm base, a positive scalar (enter 0 for natural log) 
        :Norm:  - Normalisation of ``Pnorm`` value, a boolean:
            * False -  normalises w.r.t log(# of permutation symbols [``m-1]``) - default
            *  True  -  normalises w.r.t log(# of all possible permutations [``m!``])
              * Note: Normalised permutation entropy is undefined for ``m`` = 1.
              ** Note: When ``Typex = 'uniquant'`` and ``Norm = True``, normalisation is calculated w.r.t. ``log(tpx^m)`` **
              
        :Typex: - Permutation entropy variation, one of the following: 
                {``'uniquant'``, ``'finegrain'``, ``'modified'``, ``'ampaware'``, ``'weighted'``, ``'edge'``, ``'phase'``}
                 See the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_ for more info on PermEn variants.    
        :tpx:   - Tuning parameter for associated permutation entropy variation.
                     * [uniquant]  'tpx' is the L parameter, an integer > 1 (default = 4).           
                     * [finegrain] 'tpx' is the alpha parameter, a positive scalar (default = 1)
                     * [ampaware]  'tpx' is the A parameter, a value in range [0 1] (default = 0.5)
                     * [edge]      'tpx' is the r sensitivity parameter, a scalar > 0 (default = 1)
              
        See the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_ for more info on PermEn variants.    
            
  :See also: 
      ``XPermEn``, ``MSEn``, ``XMSEn``, ``SampEn``, ``ApEn``, ``CondEn``
  
  :References:
      [1] Christoph Bandt and Bernd Pompe, 
          "Permutation entropy: A natural complexity measure for time series." 
          Physical Review Letters,
          88.17 (2002): 174102.

      [2] Xiao-Feng Liu, and Wang Yue,
          "Fine-grained permutation entropy as a measure of natural 
           complexity for time series." 
           Chinese Physics B 
           18.7 (2009): 2690.

      [3] Chunhua Bian, et al.,
          "Modified permutation-entropy analysis of heartbeat dynamics."
          Physical Review E
          85.2 (2012) : 021906

      [4] Bilal Fadlallah, et al.,
          "Weighted-permutation entropy: A complexity measure for time 
          series incorporating amplitude information." 
          Physical Review E 
          87.2 (2013): 022911.

      [5] Hamed Azami and Javier Escudero,
          "Amplitude-aware permutation entropy: Illustration in spike 
          detection and signal segmentation." 
          Computer methods and programs in biomedicine,
          128 (2016): 40-51.

      [6] Zhiqiang Huo, et al.,
          "Edge Permutation Entropy: An Improved Entropy Measure for 
          Time-Series Analysis," 
          45th Annual Conference of the IEEE Industrial Electronics Soc,
          (2019), 5998-6003

      [7] Zhe Chen, et al. 
          "Improved permutation entropy for measuring complexity of time
          series under noisy condition." 
          Complexity 
          1403829 (2019).

      [8] Maik Riedl, Andreas Müller, and Niels Wessel,
          "Practical considerations of permutation entropy." 
          The European Physical Journal Special Topics 
          222.2 (2013): 249-262.

      [9] Kang Huan, Xiaofeng Zhang, and Guangbin Zhang,
          "Phase permutation entropy: A complexity measure for nonlinear time 
          series incorporating phase information." 
          Physica A: Statistical Mechanics and its Applications 
          568 (2021): 125686.   

    """

    Sig = np.squeeze(Sig)          
    N = Sig.shape[0]
    if Logx == 0:
        Logx = np.exp(1)
    
    assert N>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 1), "m:     must be an integer > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - False or True"
    assert Typex.lower() in ['none','uniquant','finegrain','modified','ampaware','weighted','edge','phase'] \
            and isinstance(Typex,str), "Typex:    must be one of the following strings - \
            'none','uniquant','finegrain','modified','ampaware','weighted','edge', 'phase'"
    assert isinstance(tpx,(int,float)), \
            "tpx:   the value of tpx relates to 'Typex'. \
             Type help(EntropyHub.PermEn) for further info on the 'tpx' value."   
             
    if Typex.lower() == 'phase':
        Sig = np.angle(hilbert(Sig))
        if tpx == 1:
            Sig = np.unwrap(Sig)

    Sx = np.zeros((N,m))               
    Perm = np.zeros(m)
    Pnorm = np.zeros(m)
    for k in range(m):
        Nx = N-(k*tau)
        Sx[:Nx,k] = Sig[(k*tau):N]
        Temp = np.argsort(Sx[:Nx,:k+1],axis=1)
        Px = np.asarray(list(prm(np.arange(k+1))))
        Counter = np.zeros(Px.shape[0])
                
        if Typex.lower() == 'uniquant':
            Temp = np.sort(Sx[:Nx,:k+1],axis=1)
            S = np.zeros(Temp.shape)
            if tpx == -1:
                tpx = 4
            elif tpx <= 1:
                raise Exception('L parameter must be an integer > 1')
            elif not isinstance(tpx,int):
                raise Exception('L parameter must be an integer')
            
            delta = np.ptp(Sig)/tpx;
            S[:,0] = np.digitize(Temp[:,0],np.arange(min(Sig),max(Sig),delta))            
            if k > 0:
                tt = Temp[:,1:k+1] - np.transpose(np.tile(Temp[:,0],(k,1)))
                S[:,1:k+1] = np.transpose(np.tile(S[:,0],(k,1))) + np.floor(tt/delta)
                del tt

            Px = np.unique(S,axis=0)
            Counter = np.zeros(Px.shape[0])
            for n in range(len(Px)):
                Counter[n] = sum(np.any(S - Px[n,:],axis=1)==False)       
            Counter = Counter[Counter!=0]
            Ppi = Counter/sum(Counter)
            Norm = 3            
            del S, delta, n,
            
        elif Typex.lower() == 'finegrain':  
            
            if k>0:
                if tpx == -1:
                    tpx = 1
                elif tpx <= 0:
                    raise Exception('Alpha parameter must be greater than 0')
                
                q =  np.floor(np.max(abs(np.diff(Sx[:Nx,:k+1],axis=1)),axis=1)/(tpx*np.std(abs(np.diff(Sig)))))
                Temp = np.concatenate((Temp,np.expand_dims(q,1)),axis=1)
                Px = np.unique(Temp,axis=0)
                Counter = np.zeros(Px.shape[0])
                for n in range(Px.shape[0]):
                    Counter[n] = sum(np.any(Temp - Px[n,:],axis=1)==False) 
                Counter = Counter[Counter!=0]
                Ppi = Counter/sum(Counter)
                del q, n
            else:
                Ppi = 1
            
        elif Typex.lower() == 'modified':
            if k > 0:                
                Tx = np.diff(np.sort(Sx[:Nx,:k+1],axis=1),axis=1)==0    
                for km in range(k):
                    Temp[Tx[:,km],km+1] = Temp[Tx[:,km],km]                    
                #Px = np.unique(Temp,axis=0) #This is much slower than unique on it's own
                if Tx.any()>0:
                    Px = np.vstack((Px,np.unique(Temp[np.any(Tx,axis=1),:],axis=0)))
                Counter = np.zeros(Px.shape[0])
                for n in range(Px.shape[0]):
                    Counter[n] = np.sum(np.any(Temp - Px[n,:],axis=1)==False)                        
                Counter = Counter[Counter!= 0]
                Ppi = Counter/np.sum(Counter)
                del Tx
            else:
                Ppi = 1
        
        elif Typex.lower() == 'weighted':
            if k > 0:
                Wj = np.var(Sx[:Nx,:k+1],axis=1)                
                for n in range(Px.shape[0]):
                    Counter[n] = np.sum(Wj[np.any(Temp - Px[n,:],axis=1)==False])            
                Counter = Counter[Counter!= 0]
                Ppi = Counter/np.sum(Wj)
                del Wj, n
            else:
                Ppi = 1
            
        elif Typex.lower() == 'ampaware':
            if k > 0:
                if tpx ==-1:
                    tpx = 0.5
                elif tpx < 0 or tpx > 1:
                    raise Exception('The A parameter (tpx) must be in the range [0 1]')
                                      
                AA = np.sum(abs(Sx[:Nx,:k+1]),axis=1)    
                AB = np.sum(abs(np.diff(Sx[:Nx,:k+1],axis=1)),axis=1)               
                Ax = ((tpx*AA/(k+1)) + (1-tpx)*AB/(k))                
                for n in range(Px.shape[0]):
                    Counter[n] = np.sum(Ax[np.any(Temp - Px[n,:],axis=1)==False]) # Does Px need 2nd dim?
                Counter = Counter[Counter!= 0]
                Ppi = Counter/np.sum(Ax)                
                del AA, AB, Ax
            else:
                Ppi = 1 
        
        elif Typex.lower() == 'edge':
            if tpx == -1:
                tpx = 1
            elif tpx <= 0:
                raise Exception('r sensitivity parameter (tpx) must be > 0')
            
            if k > 0:
                for n in range(Px.shape[0]):
                    Sy = Sx[:Nx,:k+1]
                    Tx = np.diff(Sy[np.any(Temp - Px[n],axis=1)==False,:k+1],axis=1)
                    Counter[n] = np.sum(np.mean(np.hypot(Tx,1),axis=1)**tpx)
                Counter = Counter[Counter!= 0]
                Ppi = Counter/np.sum(Counter)
            else:
                Ppi = 1              
            
        else:
            for n in range(Px.shape[0]):
                Counter[n] = sum(np.any(Temp - Px[n,:],axis=1)==False)            
            Counter = Counter[Counter!=0]
            Ppi = Counter/sum(Counter)
        
        if np.round(np.sum(Ppi),3) != 1:
            print('Warning: Potential error with probability calculation')
               
        Perm[k] = -np.sum(Ppi*(np.log(Ppi)/np.log(Logx)))
        
        if Norm == 3:
            Pnorm[k] = Perm[k]/(np.log(tpx**(k+1))/np.log(Logx))
        
        elif Norm and k>0:
            Pnorm[k] = Perm[k]/(np.log(np.math.factorial(k+1))/np.log(Logx))
                
        elif not Norm and k>0:
            Pnorm[k] = Perm[k]/k  
            
        else:
            Pnorm[k] = np.nan
        
        del Temp, Ppi, Counter

    cPE = np.diff(Perm)
    return Perm, Pnorm, cPE
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Diversity Entropy function."""
import numpy as np 

def DivEn(Sig,  m=2, tau=1, r=5, Logx=np.exp(1)):
    """DivEn  estimates the diversity entropy of a univariate data sequence.
    
    .. code-block:: python
        
        Div, CDEn, Bm = DivEn(Sig) 
        
    Returns the diversity entropy (``Div``), the cumulative diversity entropy (``CDEn``),
    and the corresponding probabilities (``Bm``) estimated from the data sequence (``Sig``) 
    using the default parameters:   embedding dimension = 2, time delay = 1, 
    # bins = 5,  logarithm = natural,

    .. code-block:: python
        
        Div, CDEn, Bm = DivEn(Sig, keyword = value, ...)
        
    Returns the diversity entropy (``Div``) estimated from the data
    sequence (``Sig``) using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, an integer > 1
        :tau:   - Time Delay, a positive integer
        :r:     - Histogram bins #: either
                    * an integer [1 < ``r``] representing the number of bins
                    * a list/numpy array of 3 or more increasing values in range [-1 1] representing the bin edges including the rightmost edge.
        :Logx:  - Logarithm base, a positive scalar (enter 0 for natural log) 

    :See also:
        ``CoSiEn``, ``PhasEn``, ``SlopEn``, ``GridEn``, ``MSEn``
    
    :References:            
        [1] X. Wang, S. Si and Y. Li, 
            "Multiscale Diversity Entropy: A Novel Dynamical Measure for Fault 
            Diagnosis of Rotating Machinery," 
            IEEE Transactions on Industrial Informatics,
            vol. 17, no. 8, pp. 5419-5429, Aug. 2021
            
        [2] Y. Wang, M. Liu, Y. Guo, F. Shu, C. Chen and W. Chen, 
            "Cumulative Diversity Pattern Entropy (CDEn): A High-Performance, 
            Almost-Parameter-Free Complexity Estimator for Nonstationary Time Series,"
            IEEE Transactions on Industrial Informatics
            vol. 19, no. 9, pp. 9642-9653, Sept. 2023            
    """
    
    Sig = np.squeeze(Sig)
    N = Sig.shape[0]  
    if Logx == 0:
        Logx = np.exp(1)
        
    assert N>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 1), "m:     must be an integer > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert (isinstance(r, int) and r>1) or (isinstance(r, (list,np.ndarray)) and  
            len(r)>2 and min(r)>=-1 and np.ndim(r)==1 and max(r)<=1 and min(np.diff(r))>0), \
            "r:    must be an int > 1 or a list/numpy array of 3 or more increasing values in range [-1 1]"
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"
    
    Nx = N - (m-1)*tau
    Zm = np.zeros((Nx,m))
    for n in range(m):
        Zm[:,n] = Sig[n*tau:Nx+(n*tau)]
        
    Num = np.sum(Zm[:-1,:]*Zm[1:,:],axis=1)
    Den = np.sqrt(np.sum(Zm[1:,:]**2,axis=1))*np.sqrt(np.sum(Zm[:-1,:]**2,axis=1))
    Di = Num/Den
    Bm, _ = np.histogram(Di, bins = r, range=(-1,1))
    Bm = Bm[Bm>0]/sum(Bm)
    
    if np.round(np.sum(Bm),6) != 1:
        print("""Warning:
              Potential error is probability estimation!
              Sum(Pi) == """, np.round(np.sum(Bm),6))
    
    if isinstance(r, (list,np.ndarray)):     r = len(r)-1
    
    Pj = 1 - np.cumsum(Bm);     Pj = (Pj/sum(Pj))[:-1]
    CDEn = -np.sum(Pj*np.log(Pj)/np.log(Logx))/(np.log(r)/np.log(Logx))
    
    Div = -np.sum(Bm*np.log(Bm)/np.log(Logx))/(np.log(r)/np.log(Logx))
        
    return Div, CDEn, Bm                
        

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Slope Entropy function."""
import numpy as np

def SlopEn(Sig, m=2, tau=1, Lvls=(5,45), Logx=2, Norm=True):
    """SlopEn  estimates the slope entropy of a univariate data sequence.

    .. code-block:: python
    
        Slop = SlopEn(Sig) 
        
    Returns the slope entropy (``Slop``) of the data sequence (``Sig``) for embedding 
    dimensions [2, ..., ``m``] using the default parameters:  embedding dimension = 2, 
    time delay = 1, angular thresholds = [5 45],  logarithm = base 2 
  
    .. code-block:: python
    
        [Slop] = SlopEn(Sig, keyword = value, ...)
        
    Returns the slope entropy (``Slop``) estimates of the data sequence (``Sig``)  
    using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, an integer > 1
              SlopEn returns estimates for each dimension [2,..., ``m``]
        :tau:   - Time Delay, a positive integer
        :Lvls:  - Angular thresolds, a vector of monotonically increasing  values in the range [0 90] degrees.
        :Logx:  - Logarithm base, a positive scalar (enter 0 for natural log)
        :Norm:  - Normalisation of ``SlopEn`` value, one of the following integers:
            * [False]  no normalisation
            * [True]   normalises w.r.t. the number of patterns found (default)
 
    :See also:
        ``PhasEn``, ``GridEn``, ``MSEn``, ``CoSiEn``, ``SampEn``, ``ApEn``
  
    :References:
        [1] David Cuesta-Frau,
            "Slope Entropy: A New Time Series Complexity Estimator Based on
            Both Symbolic Patterns and Amplitude Information." 
            Entropy 
            21.12 (2019): 1167.
    """
    
    Sig = np.squeeze(Sig)
    if Logx == 0:
        Logx = np.exp(1)
    assert Sig.shape[0]>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 1), "m:     must be an integer > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - True or False"  
    assert isinstance(Lvls,(tuple,np.ndarray)) and len(Lvls)>1 and min(np.diff(Lvls))>0 \
        and  min(Lvls)>0 and max(Lvls)<=90, "Lvls:  must be a tuple/numpy \
        vector with multiple monotonically increasing values in range [0 90]"
               
    m = m-1
    Tx = np.degrees(np.arctan(Sig[tau:] - Sig[:-tau]))
    N = Tx.shape[0]
    Sx = np.zeros((N,m))
    Symbx = np.zeros(N) #np.shape(Tx))
    Slop = np.zeros(m)    
    for q in range(1,len(Lvls)):        
        Symbx[np.logical_and(Tx<= Lvls[q], Tx> Lvls[q-1])] = q
        Symbx[np.logical_and(Tx>=-Lvls[q], Tx<-Lvls[q-1])] = -q
        
        if q == len(Lvls)-1:
            Symbx[Tx> Lvls[q]] = q+1
            Symbx[Tx<-Lvls[q]] = -(q+1)
               
    for k in range(m):
        Sx[:N-k,k] = Symbx[k:N]
        _,Locs = np.unique(Sx[:N-k,:k+1],axis=0,return_counts=True)
        
        if Norm:
            p = Locs/(N-k)        
            if np.round(np.sum(p))!= 1:
                print('WARNING: Potential Error: Some permutations not accounted for!')
        else:
            p = Locs/len(Locs)
        
        Slop[k] = -np.sum(p*np.log(p)/np.log(Logx))
        del Locs, p
    
    return Slop
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Spectral Entropy function."""
import numpy as np  

def SpecEn(Sig, N=None, Freqs=(0,1), Logx=np.exp(1), Norm=True):
    """SpecEn  estimates the spectral entropy of a univariate data sequence.

    .. code-block:: python
    
        Spec, BandEn = SpecEn(Sig) 
    
    Returns the spectral entropy estimate of the full spectrum (``Spec``)
    and the within-band entropy (BandEn) estimated from the data sequence (``Sig``)
    using the default  parameters: 
    N-point FFT = 2*len(Sig) + 1, normalised band edge frequencies = [0 1],
    logarithm = base 2, normalisation = w.r.t # of spectrum/band frequency values.
 
    .. code-block:: python
    
        Spec, BandEn = SpecEn(Sig, keyword = value, ...)
        
    Returns the spectral entropy (``Spec``) and the within-band entropy (``BandEn``)
    estimate from the data sequence (``Sig``) using the specified 'keyword' arguments:
        :N:     - Resolution of spectrum (N-point FFT), an integer > 1
        :Freqs: - Normalised band edge frequencies, a 2 element tuple with values in range [0 1] where 1 corresponds to the Nyquist frequency (Fs/2).
             * Note: When no band frequencies are entered, ``BandEn == SpecEn``
        :Logx:  - Logarithm base, a positive scalar (enter 0 for natural log) 
        :Norm:  - Normalisation of Spec value, a boolean:
           * [False]  no normalisation.
           * [True]   normalises w.r.t # of spectrum/band frequency values - default.
 
    For more info, see the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_.
    
    :See also:
        ``XSpecEn``, ``MSEn``, ``numpy.fftpack``
          
    :References:
        [1] G.E. Powell and I.C. Percival,
            "A spectral entropy method for distinguishing regular and 
            irregular motion of Hamiltonian systems." 
            Journal of Physics A: Mathematical and General 
            12.11 (1979): 2053.
    
        [2] Tsuyoshi Inouye, et al.,
            "Quantification of EEG irregularity by use of the entropy of 
            the power spectrum." 
            Electroencephalography and clinical neurophysiology 
            79.3 (1991): 204-210.
    
    """
  
    Sig = np.squeeze(Sig)            
    if N is None:
        N = 2*len(np.squeeze(Sig)) + 1
    
    assert Sig.shape[0]>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert N>1 and isinstance(N,int), "N:   must be an integer > 1"
    assert isinstance(Logx,(int,float)) and Logx > 0, "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - True or False"       
    assert isinstance(Freqs,tuple) and len(Freqs)==2 and Freqs[0]>=0 and Freqs[1]<=1,\
            "Freq:    must be a two element tuple with values in range [0 1]. \
                      The values must be in increasing order."  
        
    Fx = int(np.ceil(N/2))
    Freqs = np.round(np.array(Freqs)*Fx).astype(int)-1
    Freqs[Freqs==-1] = 0    
    if Freqs[0] > Freqs[1]:
        raise Exception('Lower band frequency must come first.')
    elif Freqs[1] - Freqs[0] < 1:
        raise Exception('Spectrum resoution too low to determine bandwidth.') 
    elif min(Freqs)<0 or max(Freqs)>Fx:
        raise Exception('Freqs must be normalized w.r.t. sampling frequency [0 1].')
        
    Pt = abs(np.fft.fft(np.convolve(Sig,Sig),N))
    Pxx = Pt[:Fx]/sum(Pt[:Fx])
    Spec = -sum(Pxx*np.log(Pxx))/np.log(Logx)
    Pband = Pxx[Freqs[0]:Freqs[1]+1]/sum(Pxx[Freqs[0]:Freqs[1]+1])
    BandEn = -sum(Pband*np.log(Pband))/np.log(Logx)
    
    if Norm:
        Spec = Spec/(np.log(Fx)/np.log(Logx))
        BandEn = BandEn/(np.log(np.diff(Freqs)[0]+1)/np.log(Logx))
    return Spec, BandEn
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"Base Composite / Refined-Composite Multivariate Multiscale Entropy function."
import numpy as np   
from matplotlib.pyplot import figure, axes, show
from copy import deepcopy
# from scipy.stats import moment

def cMvMSEn(Data, Mbjx, Scales=3, Refined=False, Plotx=False):
    """cMvMSEn  Returns the composite + refined-composite multivariate multiscale entropy of a multivariate dataset.
    
    .. code-block:: python
    
        MSx,CI = cMvMSEn(Data, Mobj) 
    
    Returns a vector of composite multivariate multiscale entropy values (``MSx``) and the complexity 
    index (``CI``) of the data sequences in ``Data`` using the parameters specified 
    by the multiscale object (``Mobj``) over 3 temporal scales with coarse-
    graining (default). 
    
    .. caution::
 
        By default, the ``MvSampEn`` and ``MvFuzzEn`` multivariate entropy algorithms
        estimate entropy values using the "full"  method by comparing delay vectors 
        across all possible ``m+1`` expansions of the embedding space as applied in [1].
        These methods are **not** lower-bounded to 0, like most entropy algorithms,
        so ``MvMSEn`` may return negative entropy values if the base multivariate 
        entropy function is ``MvSampEn`` and ``MvFuzzEn``, even for stochastic processes...
    
    

    .. code-block:: python 
        
        MSx, CI = cMvMSEn(Data, Mobj, Refined = True) 
        
    Returns a vector of refined-composite multiscale entropy values (``MSx``) for the data 
    sequences in (``Data``) using the parameters specified by the multiscale object 
    (``Mobj``) using the refined-composite multivariate multiscale entropy method (rcMSE) over 3 temporal
    scales. When ``Refined == True``, the base entropy method must be ``MvSampEn`` or ``MvFuzzEn``.
    If the entropy method is ``MvSampEn``, cMvMSEn employs the method described in [1]. 
    If the entropy method is ``MvFuzzEn``, cMvMSEn employs the method described in [5]. 

    .. code-block:: python
    
        MSx,CI = cMvMSEn(Data, Mobj, keyword = value, ...)
        
    Returns a vector of composite multivariate multiscale entropy values (``MSx``) and the complexity 
    index (``CI``) of the data sequences in ``Data`` using the parameters specified by
    the multiscale object (``Mobj``) and the following ``keyword`` arguments:
        :Scales:   - Number of temporal scales, an integer > 1   (default: 3)
        :Refined:  - Refined-composite MvMSEn method. When ``Refined == True`` and the entropy function specified by ``Mobj`` is ``MvSampEn`` or ``MvFuzzEn``, ``cMvMSEn`` returns the refined-composite multivariate multiscale entropy (rcMSEn) [default: False]
        :Plotx:    - When ``Plotx == True``, returns a plot of the entropy value at each time scale (i.e. the multiscale entropy curve) [default: False]
    
    For further info on these graining procedures see the `EntropyHub guide <https://github.com/MattWillFlood/EntropyHub/blob/main/EntropyHub%20Guide.pdf>`_.
    
    :See also:
        ``MSobject``, ``MvMSEn``, ``MvFuzzEn``, ``MvSampEn``, ``MvPermEn``, ``MvCoSiEn``,  ``MvDispEn``
    
    :References:
        [1] Shuen-De Wu, et al.,
              "Time series analysis using composite multiscale entropy."
              Entropy
              15.3 (2013): 1069-1084.
    
        [2] Shuen-De Wu, et al.,
              "Analysis of complex time series using refined composite
              multiscale entropy."
              Physics Letters A
              378.20 (2014): 1369-1374.
    
        [3] Ahmed Mosabber Uddin, Danilo P. Mandic
              "Multivariate multiscale entropy: A tool for complexity
              analysis of multichannel data."
              Physical Review E 84.6 (2011): 061918.
    
        [4] Ahmed Mosabber Uddin, Danilo P. Mandic
              "Multivariate multiscale entropy analysis."
              IEEE signal processing letters 19.2 (2011): 91-94.
    
        [5] Azami, Alberto Fernández, Javier Escudero.
              "Refined multiscale fuzzy entropy based on standard deviation for
              biomedical signal analysis."
              Medical & biological engineering & computing 55 (2017): 2037-2052.
    
    """
    
    Mobj = deepcopy(Mbjx)    
    Data = np.squeeze(Data)
    assert Data.shape[0]>10 and Data.ndim==2 and Data.shape[1]>1,  "Data:   must be an NxM numpy matrix where N>10 and M>1"
    N, Dn = Data.shape     
    
    assert isinstance(Mobj,object), "Mobj:  must be a multiscale entropy \
                        object created with the function EntropyHub.MSobject"    
    assert isinstance(Scales,int) and Scales>1, "Scales:    must be an integer > 1"
    assert isinstance(Refined, bool) and ((Refined==True and Mobj.Func.__name__ in ['MvSampEn','MvFuzzEn'])
        or Refined==False), \
    "Refined:       must be a bool (True or False). If Refined==True, Mobj.Func must be MvSampEn or MvFuzzEn" 
    assert isinstance(Plotx, bool), "Plotx:    must be boolean - True or False"
    assert Mobj.Func.__name__[:2] == 'Mv', ("Base entropy estimator must be a multivariate entropy method." 
    "To perform univariate multiscale entropy estimation, use MSEn().")
           
    if Refined:
        if  Mobj.Func.__name__== 'MvFuzzEn':      Tx = 1
        elif Mobj.Func.__name__ == 'MvSampEn':    Tx = 0
        
        if 'Logx' in Mobj.Kwargs.keys():   Logx = Mobj.Kwargs['Logx']
        else:   Logx = np.exp(1)
        
    else: 
        Tx = 0
  
    MSx = np.zeros(Scales)    
    for T in range(1,Scales+1):
        print(' .', end='')
        Temp = modified(Data, T, Tx, Dn)          
        N = T*(Temp.shape[0]//T)
        Ma = np.zeros(T)   
        Mb = np.zeros(T)     

        for k in range(T):        
            if Refined:  _, Ma[k], Mb[k], _ = Mobj.Func(Temp[k:N:T,:],**Mobj.Kwargs)
            else:   Ma[k] = Mobj.Func(Temp[k:N:T,:],**Mobj.Kwargs)[0]
    
        if Refined:
            MSx[T-1] = -np.log(np.sum(Mb)/np.sum(Ma))/np.log(Logx)
        else:
            MSx[T-1] = np.mean(Ma)       
                   
    CI = sum(MSx)
    if np.any(np.isnan(MSx)):
        print('Some entropy values may be undefined.')

    if Plotx:
        if Refined:
            strx = 'Refined-Composite'
        else:
            strx = 'Composite'
        
        figure()
        ax1 = axes()   
        ax1.plot(np.arange(1,Scales+1), MSx, color=(8/255, 63/255, 77/255), linewidth=3)
        ax1.scatter(np.arange(1,Scales+1), MSx, 60, color=(1,0,1))
        ax1.set_xlabel('Scale Factor',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
        ax1.set_ylabel('Entropy Value',fontsize=12,fontweight='bold',color=(7/255, 54/255, 66/255))
        ax1.set_title('%s Multivariate Multiscale %s'%(strx,Mobj.Func.__name__[2:]), 
                     fontsize=16,fontweight='bold',color=(7/255, 54/255, 66/255))       
        show()
    
    return MSx, CI


def modified(Z, sx, Tx, Dn):    
    Ns = Z.shape[0]-sx+1
    Y = np.zeros((Ns,Dn))    
    if Tx==1:
        for k in range(Dn):
            Temp = [np.std(Z[x:x+sx,k]) for x in range(Ns)]
            Y[:,k] = np.array(Temp)
    
    else:
        for k in range(Dn):
            Y[:,k] = np.convolve(Z[:,k],np.ones(sx),'valid')/sx    
   
    return Y
    

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Entropy of Entropy Function"""    
import numpy as np

def EnofEn(Sig, tau=10, S=10, Xrange=None, Logx=np.exp(1)):
    """EnofEn  estimates the entropy of entropy of a univariate data sequence.
    
    .. code-block:: python

        EoE, AvEn, S2 = EnofEn(Sig) 

    Returns the entropy of entropy (``EoE``), the average Shannon entropy 
    across all windows (``AvEn``), and the number of levels (``S2``) estimated 
    from the data sequence (``Sig``) using  the default parameters: 
    window length (samples) = 10, slices = 10, logarithm = natural
    heartbeat interval range (xmin, xmax) = (min(Sig), max(Sig))

    .. code-block:: python
        
        EoE, AvEn, S2 = EnofEn(Sig, keyword = value, ...)
        
    Returns the entropy of entropy (``EoE``) estimated from the data sequence (``Sig``)  
    using the specified 'keyword' arguments:
        :tau:     - Window length, an integer > 1
        :S:       - Number of slices, an integer > 1
        :Xrange:  - The min and max heartbeat interval, a two-element tuple where X[0] < X[1]
        :Logx:    - Logarithm base, a positive scalar  
 
    :See also::
        ``SampEn``, ``MSEn``
    
    :References:
        [1] Chang Francis Hsu, et al.,
            "Entropy of entropy: Measurement of dynamical complexity for biological systems." 
            Entropy 
            19.10 (2017): 550.
    
    """
        
    Sig = np.squeeze(Sig)
    
    if Xrange is None:
        Xrange = (np.min(Sig),np.max(Sig))

    assert Sig.shape[0]>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"
    assert isinstance(tau,int) and (tau > 1) and (tau<len(Sig)), "tau:   must be an integer > 1 and < length(Sig)"
    assert isinstance(S,int) and S>1, "S:  must be an integer > 1"
    assert isinstance(Xrange,tuple) and Xrange[0]<=Xrange[1] and len(Xrange)==2, \
    "Xrange:  must be a two-element numeric tuple where Xrange[0]<Xrange[1]"   
    
    Wn = int(np.floor(Sig.shape[0]/tau))
    Wj = np.reshape(Sig[:Wn*tau],(Wn,tau))
    Edges = np.linspace(Xrange[0],Xrange[1],S+1)   #Edges = np.linspace(min(Sig),max(Sig),S[0]+1)
    Yj = np.zeros(Wn)
    for n in range(Wn):
        Temp,_ = np.histogram(Wj[n,:],Edges)        
        Temp = Temp[Temp!=0]/tau
        Yj[n] = -sum(Temp*(np.log(Temp)/np.log(Logx)))
        
    AvEn = sum(Yj)/Wn
    #Edges = np.linspace(min(Yj),max(Yj),S[1]+1)
    #Pjl,_ = np.histogram(Yj,Edges)
    #Pjl = Pjl/Wn
    #Pjl = Pjl[Pjl!=0]
    
    _, Tempy = np.unique(np.round(Yj,12), return_counts=True)    
    Pjl = Tempy/Wn
    S2 = len(Pjl)
    
    if round(sum(Pjl),5) != 1:
        print('Warning: Possible error estimating probabilities')
    EoE = -sum(Pjl*(np.log(Pjl)/np.log(Logx)))    
        
    return EoE, AvEn, S2
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


""" Attention Entropy function."""
import numpy as np

def AttnEn(Sig, Logx=2):
    """AttnEn  estimates the attention entropy of a univariate data sequence.
    
    .. code-block:: python
    
        Attn, (Hxx, Hnn, Hxn, Hnx) = AttnEn(Sig)
    
    Returns the attention entropy (``Attn``) calculated as the average of the
    sub-entropies (``Hxx``, ``Hxn``, ``Hnn``, ``Hnx``) estimated from the data sequence (``Sig``) 
    using a base-2 logarithm.
    
    .. code-block:: python
    
        Attn, (Hxx, Hnn, Hxn, Hnx) = AttnEn(Sig, Logx = value)
    
    Returns the attention entropy (``Attn``) and a four-element tuple of
    sub-entropies (``Hxx``, ``Hnn``, ``Hxn``, ``Hnx``) from the data sequence (``Sig``) where,
        :Hxx:  -  entropy of local-maxima intervals
        :Hnn: -   entropy of local minima intervals
        :Hxn: -   entropy of intervals between local maxima and subsequent minima
        :Hnx:   - entropy of intervals between local minima and subsequent maxima using the following keyword argument:
        :Logx:  - Logarithm base, a positive scalar  (enter 0 for natural log)
    
    :See also:
        ``EnofEn``, ``SpecEn``, ``XSpecEn``, ``PermEn``, ``MSEn``
    
    :References:
        [1] Jiawei Yang, et al.,
            "Classification of Interbeat Interval Time-series Using Attention Entropy." 
            IEEE Transactions on Affective Computing 
            (2020)
    """

    Sig = np.squeeze(Sig)
    N = Sig.shape[0]
    if Logx == 0:
        Logx = np.exp(1) 
    assert N>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"

    Xmax = PkFind(Sig)
    Xmin = PkFind(-Sig)
    Txx = np.diff(Xmax)
    Tnn = np.diff(Xmin)
    Temp = np.diff(np.sort(np.hstack((Xmax, Xmin))))
    
    assert len(Xmax) > 0, "No local maxima found!"
    assert len(Xmin) > 0, "No local minima found!" 
    
    if Xmax[0]<Xmin[0]:
       Txn = Temp[::2]
       Tnx = Temp[1::2]
    else:
       Txn = Temp[1::2]
       Tnx = Temp[::2]
       
    Edges = np.arange(-0.5,N+1)
    Pnx,_ = np.histogram(Tnx,Edges)
    Pnn,_ = np.histogram(Tnn,Edges)
    Pxx,_ = np.histogram(Txx,Edges)
    Pxn,_ = np.histogram(Txn,Edges)
    
    Pnx = Pnx[Pnx!=0]/len(Tnx)
    Pxn = Pxn[Pxn!=0]/len(Txn)
    Pnn = Pnn[Pnn!=0]/len(Tnn)
    Pxx = Pxx[Pxx!=0]/len(Txx)
    
    Hxx = -sum(Pxx*np.log(Pxx)/np.log(Logx))
    Hxn = -sum(Pxn*np.log(Pxn)/np.log(Logx))
    Hnx = -sum(Pnx*np.log(Pnx)/np.log(Logx))
    Hnn = -sum(Pnn*np.log(Pnn)/np.log(Logx))
    Av4 = (Hnn + Hxx + Hxn + Hnx)/4
    
    return Av4, (Hxx,Hnn,Hxn,Hnx)


def PkFind(X):
    Nx = len(X)
    Indx = np.zeros(Nx)
    for n in range(1,Nx-1):
        if X[n-1] < X[n] > X[n+1]:
            Indx[n] = n

        elif X[n-1] < X[n] == X[n+1]:
            k = 1
            while (n+k)<Nx-1 and X[n] == X[n+k]:
                k += 1
            
            if X[n] > X[n+k]:
                Indx[n] = n + ((k-1)//2)

    Indx = Indx[Indx!=0]
    return Indx

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Bidimensional Permutation Entropy function."""
import numpy as np   

def PermEn2D(Mat, m=None, tau=1, Norm=True, Logx=np.exp(1), Lock=True):
    """PermEn2D  Estimates the bidimensional permutation entropy of a data matrix.

    .. code-block:: python
    
        Perm2D = PermEn2D(Mat) 
            
    Returns the bidimensional permutation entropy estimate (``Perm2D``) estimated for 
    the data matrix (``Mat``) using the default parameters: time delay = 1,
    logarithm = natural, template matrix size = [floor(H/10) floor(W/10)]  
    (where H and W represent the height (rows) and width (columns) of the data matrix ``Mat``) 
    ** The minimum number of rows and columns of ``Mat`` must be > 10.
 
    .. code-block:: python
     
         Perm2D = PermEn2D(Mat, keyword = value, ...)
    
    Returns the bidimensional permutation entropy (``Perm2D``) estimates for the data
    matrix (``Mat``) using the specified 'keyword' arguments:
        :m:     - Template submatrix dimensions, an integer scalar (for sub-matrix with same height and width) or a two-element tuple of integers (height, width) with a minimum value > 1. (default: [floor(H/10) floor(W/10)])
        :tau:   - Time Delay, a positive integer > 1   (default: 1)
        :Norm:  - Normalization of the PermEn2D value by maximum Shannon entropy (Smax = log((mx*my)!)         [default: true]
        :Logx:  - Logarithm base, a positive scalar    (default: natural)        
        :Lock:  - By default, ``PermEn2D`` only permits matrices with a maximum size of 128 x 128 to prevent RAM overload. 
                    e.g. For ``Mat`` = [200 x 200], ``m = 3``, and ``tau = 1``, ``PermEn2D`` 
                    creates a vector of 753049836 elements. To enable matrices
                    greater than [128 x 128] elements, set ``'Lock' = false``. (default: true)
                    ``CAUTION: unlocking the permitted matrix size may cause memory``
                    ``errors that could lead your Python IDE to crash.``
                    
        
    .. note:: 

	     **The original bidimensional permutation entropy algorithms [1][2]**
	     **do not account for equal-valued elements of the embedding matrices.**
             To overcome this, PermEn2D uses the lowest common rank for
             such instances. For example, given an embedding matrix A where,

             :A: 
		|	[3.4  5.5  7.3]         
                |   	[2.1  6    9.9]       
                |   	[7.3  1.1  2.1]      
  
             would normally be mapped to an ordinal pattern like so,    
  
                |       [3.4  5.5  7.3  2.1  6  9.9  7.3  1.1  2.1] =>
                |       [ 8    4    9    1   2   5    3    7    6 ]

             However, indices 4 & 9, and 3 & 7 have the same values, 2.1
             and 7.3 respectively. Instead, PermEn2D uses the ordinal pattern
             [ 8    4    4    1   2   5    3    3    6 ]
             where the lowest rank (4 & 3) are used instead (of 9 & 7). 
             Therefore, the number of possible permutations is no longer 
             (mx*my)!, but (mx*my)^(mx*my). Here, the PermEn2D value is 
             normalized by the maximum Shannon entropy (Smax = log((mx*my)!) 
             ``assuming that no equal values are found in the permutation
             motif matrices``, as presented in [1].

    :See also:
        ``SampEn2D``, ``FuzzEn2D``, ``DistEn2D``, ``DispEn2D``, ``EspEn2D``
     
    
    :References:
        [1] Haroldo Ribeiro et al.,
            "Complexity-Entropy Causality Plane as a Complexity Measure 
            for Two-Dimensional Patterns"
            PLoS ONE (2012), 7(8):e40689, 
    
        [2] Luciano Zunino and Haroldo Ribeiro,
             "Discriminating image textures with the multiscale
             two-dimensional complexity-entropy causality plane"
             Chaos, Solitons and Fractals,  91:679-688 (2016)
    
        [3] Matthew Flood and Bernd Grimm,
             "EntropyHub: An Open-source Toolkit for Entropic Time Series Analysis"
             PLoS ONE (2021) 16(11): e0259448.
                                 
    """
            
    Mat = np.squeeze(Mat)
    NL,NW = Mat.shape    
    if m is None:
        m = np.array(Mat.shape)//10
           
    assert Mat.ndim==2 and min(Mat.shape)>10 , \
    "Mat:   must be a 2D numpy array with height & width > 10"
    assert (isinstance(m,int) and m>1) or (isinstance(m, (np.ndarray,tuple))  
                                           and len(m)==2 and min(m)>1), \
    "m:     must be an integer > 1, or a 2 element tuple of integers > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(Norm,bool), "Norm: must be a boolean"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Lock,bool) and ((Lock==True and max(Mat.shape)<=128) or Lock==False), \
    "Lock:      To prevent memory storage errors, matrix width & length must \
    have <= 128 elements. To estimate PermEn2D for the current matrix (%d x %d) \
    change Lock to False. \
    CAUTION: unlocking the permitted matrix size may cause memory \
    errors that could lead your Python IDE to crash.."%(NW,NL)
    
    if isinstance(m,int):
        mL = int(m); mW = int(m)        
    else:
        mL = int(m[0]); mW = int(m[1])       
        
    NL = NL - (mL-1)*tau
    NW = NW - (mW-1)*tau
    
    Temp = Mat[:(mL-1)*tau+1:tau,:(mW-1)*tau+1:tau]
    Sord = np.sort(Temp.flatten('F'))
    Dict = np.argsort(Temp.flatten('F'),kind='stable')
    if np.any(np.diff(Sord)==0):
        for x in np.where(np.diff(Sord)==0)[0]+1:
            Dict[x] = Dict[x-1]  
 
    Counter = 0
    Dict = np.expand_dims(Dict,axis=0)
    
    for k in range(NL):        
        for n in range(NW):            
            Temp = Mat[k:(mL-1)*tau+k+1:tau,n:(mW-1)*tau+n+1:tau]
            Sord = np.sort(Temp.flatten('F'))
            Dx = np.argsort(Temp.flatten('F'),kind='stable')
            
            if np.any(np.diff(Sord)==0):
                for x in np.where(np.diff(Sord)==0)[0]+1:
                    Dx[x] = Dx[x-1] 
                    
            if np.any(~np.any(Dict - Dx, axis=1)):
                Counter += ~np.any(Dict - Dx, axis=1)*1
                
            else:
                Dict = np.vstack((Dict, Dx))
                Counter = np.hstack((Counter, 1))
                
    if np.sum(Counter) != NL*NW:
        print('Warning: Potential error with permutation comparisons.')        

    Pi = Counter/sum(Counter) 
    Perm2D = -np.sum(Pi*np.log(Pi)/np.log(Logx))
                  
    if Norm:
        Perm2D = Perm2D/(np.log(float(np.math.factorial(mL*mW)))/np.log(Logx))
                     
    return Perm2D

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Gridded Distribution Entropy function."""
import numpy as np     
import matplotlib.pyplot as mpl

def GridEn(Sig, m=3, tau=1, Logx=np.exp(1), Plotx=False):
    """GridEn  estimates the gridded distribution entropy of a univariate data sequence.
 
    .. code-block:: python
    
        GDE, GDR, _ = GridEn(Sig) 
        
    Returns the gridded distribution entropy (``GDE``) and the gridded 
    distribution rate (``GDR``) estimated from the data sequence (``Sig``) using 
    the default  parameters: grid coarse-grain = 3, time delay = 1, logarithm = base 2
   
    .. code-block:: python
    
        GDE, GDR, PIx, GIx, SIx, AIx = GridEn(Sig, keyword = value)
    
    Returns the gridded distribution entropy (``GDE``) estimated from the data 
    sequence (``Sig``) using the specified 'keyword' arguments:
        :m:     - Grid coarse-grain (m x m sectors), an integer > 1
        :tau:   - Time Delay, a positive integer
        :Logx:  - Logarithm base, a positive scalar
        :Plotx: - When ``Plotx == True``, returns gridded Poicaré plot and a bivariate histogram of the grid point distribution (default: False)  
    
    :See also:
        ``PhasEn``, ``CoSiEn``, ``SlopEn``, ``BubbEn``, ``MSEn``
    
    :References:
        [1] Chang Yan, et al.,
            "Novel gridded descriptors of poincaré plot for analyzing 
            heartbeat interval time-series." 
            Computers in biology and medicine 
            109 (2019): 280-289.
    
        [2] Chang Yan, et al. 
            "Area asymmetry of heart rate variability signal." 
            Biomedical engineering online 
            16.1 (2017): 1-14.
    
        [3] Alberto Porta, et al.,
            "Temporal asymmetries of short-term heart period variability 
            are linked to autonomic regulation." 
            American Journal of Physiology-Regulatory, Integrative and 
            Comparative Physiology 
            295.2 (2008): R550-R557.
    
        [4] C.K. Karmakar, A.H. Khandoker and M. Palaniswami,
            "Phase asymmetry of heart rate variability signal." 
            Physiological measurement 
            36.2 (2015): 303.

    """
 
    Sig = np.squeeze(Sig)  
       
    assert Sig.shape[0]>10 and Sig.ndim == 1, "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 1), "m:     must be an integer > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"
    assert isinstance(Plotx,bool), "Plotx:     must be boolean - True or False"                          
        
    Sig_n = (Sig-min(Sig))/np.ptp(Sig)
    Temp = np.array([Sig_n[:-tau], Sig_n[tau:]])
    N, _, _ = np.histogram2d(Temp[0,:], Temp[1,:],m)
    Pj = np.flipud(np.transpose(N))/Temp.shape[1]
    Ppi = Pj[Pj>0]               
    if np.round(np.sum(Ppi)) != 1:
        print('WARNING: Potential error of estimated probabilities: P = %d'%np.sum(Ppi))
    
    GDE= -np.sum(Ppi*(np.log(Ppi)/np.log(Logx)))
    GDR = np.sum(np.sum(N!=0))/(m*m)
    
    with np.errstate(divide='ignore', invalid='ignore'):
        T2   = np.transpose(np.arctan(Temp[1,:]/Temp[0,:])*180/np.pi)
        Dup  = np.sum(abs(np.diff(Temp[:,T2>45],axis=0)))
        Dtot = np.sum(abs(np.diff(Temp[:,T2!=45],axis=0)))
        Sup  = np.sum((T2[T2>45]-45))
        Stot = np.sum(abs(T2[T2!=45]-45))
        Aup  = np.sum(abs(np.transpose(((T2[T2>45]-45))*np.sqrt(np.sum(Temp[:,T2>45]**2,axis=0)))))
        Atot = np.sum(abs(np.transpose(((T2[T2!=45]-45))*np.sqrt(np.sum(Temp[:,T2!=45]**2,axis=0)))))
        
        Index = {}
        Index['PIx'] = 100*sum(T2 < 45)/sum(T2!=45)
        Index['GIx'] = 100*Dup/Dtot
        Index['SIx'] = 100*Sup/Stot
        Index['AIx'] = 100*Aup/Atot

    if Plotx:

        ntrvl = np.linspace(0,1,m+1)
        mpl.subplots(1,2)
        x1 = mpl.subplot(121)
        ax1 =mpl.axes(x1)   
        
        ax1.plot(Sig_n[:-tau],Sig_n[tau:],'.m')
        ax1.plot(np.tile(ntrvl,(2,1)),np.array((np.zeros(m+1),np.ones(m+1))),'c')
        ax1.plot(np.array((np.zeros(m+1),np.ones(m+1))),np.tile(ntrvl,(2,1)),'c')
        ax1.plot([0, 1],[0, 1],'k') 
        ax1.set_aspect('equal', 'box')
        ax1.set_xlabel('X m')
        ax1.set_ylabel(r'$X m + \tau$') 
        ax1.set_xticks([0, 1])
        ax1.set_yticks([0, 1])
        ax1.set_xlim([0,1])
        ax1.set_ylim([0,1])

        x2 = mpl.subplot(122)
        ax2 = mpl.axes(x2)   
        # ax2.imshow(np.flipud(N), cmap='cool', aspect='equal')
        ax2.imshow(np.fliplr(N), cmap='cool', aspect='equal')
        ax2.set_xlabel('X m')
        ax2.set_ylabel(r'$X m + \tau$')
        ax2.set_xticks([])
        ax2.set_yticks([])
        mpl.show()
    
    return GDE, GDR, Index
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Increment Entropy function."""
import numpy as np

def IncrEn(Sig,  m=2, tau=1, R=4, Logx=2, Norm=False):
    """IncrEn  estimates the increment entropy of a univariate data sequence.
    
    .. code-block:: python
    
        Incr = IncrEn(Sig) 
        
    Returns the increment entropy (``Incr``) estimate of the data sequence (``Sig``)
    using the default parameters: embedding dimension = 2, time delay = 1, quantifying resolution = 4, logarithm = base 2,
 
    .. code-block:: python    
    
        Incr = IncrEn(Sig, keyword = value, ...)
    
    Returns the increment entropy (``Incr``) estimate of the data sequence (``Sig``)
    using the specified 'keyword' arguments:
        :m:     - Embedding Dimension, an integer > 1
        :tau:   - Time Delay, a positive integer
        :R:     - Quantifying resolution, a positive scalar
        :Logx:  - Logarithm base, a positive scalar (enter 0 for natural log) 
        :Norm:  - Normalisation of ``IncrEn`` value, a boolean:
            * [False]  no normalisation - default
            * [True]   normalises w.r.t embedding dimension (m-1). 
 
    :See also:
        ``PermEn``, ``SyDyEn``, ``MSEn``
    
    :References:
        [1] Xiaofeng Liu, et al.,
            "Increment entropy as a measure of complexity for time series."
            Entropy
            18.1 (2016): 22.1.
    
        ***   "Correction on Liu, X.; Jiang, A.; Xu, N.; Xue, J. - Increment 
            Entropy as a Measure of Complexity for Time Series, Entropy 2016, 18, 22." 
            Entropy 
            18.4 (2016): 133.
    
        [2] Xiaofeng Liu, et al.,
            "Appropriate use of the increment entropy for 
            electrophysiological time series." 
            Computers in biology and medicine 
            95 (2018): 13-23.
    
    """
    
    Sig = np.squeeze(Sig)
    if Logx == 0:
        Logx = np.exp(1)
    
    assert Sig.shape[0]>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 1), "m:     must be an integer > 1"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(R,int) and R > 0, "R:    must be a positive integer"
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"
    assert isinstance(Norm,bool), "Norm:     must be boolean - True or False"  
                
    Vi = np.diff(np.squeeze(Sig))
    N = Vi.shape[0]  - ((m-1)*tau)
    Vk = np.zeros((N,m))               
    for k in range(m):
        Vk[:,k] = Vi[k*tau:N+(k*tau)]        
        
    Sk = np.sign(Vk)
    Temp = np.tile(np.std(Vk,axis=1,ddof=1,keepdims=True),(1,m))
    Qk = np.minimum(R, np.floor(abs(Vk)*R/Temp))
    Qk[np.any(Temp==0,axis=1),:] = 0
    Wk = Sk*Qk
    Px = np.unique(Wk,axis=0)
    Counter = np.zeros(Px.shape[0])
    
    for k in range(Px.shape[0]):
        Counter[k] = np.sum(np.any(Wk - Px[k,:],axis=1)==False)
    Ppi = Counter/N      
        
    if Px.shape[0] > (2*R + 1)**m:
        print('Possible error with probability estimation')
    elif round(sum(Ppi),3) != 1:
        print('Possible error with probability estimation')
        
    Incr = -sum(Ppi*(np.log(Ppi)/np.log(Logx)))
    if Norm:
        Incr = Incr/(m-1)
    
    return Incr
"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


import numpy as np       
""" Base Sample Entropy function."""

def SampEn(Sig, m=2, tau=1, r=None, Logx=np.exp(1), Vcp=False):
    """SampEn  estimates the sample entropy of a univariate data sequence.

    .. code-block:: python
    
        Samp, A, B = SampEn(Sig) 
        
    Returns the sample entropy estimates (``Samp``) and the number of matched state
    vectors (``m: B``, ``m+1: A``) for ``m`` = [0, 1, 2] estimated from the data sequence (``Sig``)
    using the default parameters: embedding dimension = 2, time delay = 1, 
    radius threshold = 0.2*SD(``Sig``), logarithm = natural.    
    
    .. code-block:: python
    
        Samp, A, B, (Vcp, Ka, Kb) = SampEn(Sig, ..., Vcp = True) 
    
    If ``Vcp == True``, an additional tuple ``(Vcp, Ka, Kb)`` is returned with    
    the sample entropy estimates (``Samp``) and the number of matched state
    vectors (``m: B``, ``m+1: A``). ``(Vcp, Ka, Kb)``  contains the variance of the conditional
    probabilities (``Vcp``), i.e. CP = A/B,  and the number of **overlapping**
    matching vector pairs of lengths m+1 (``Ka``) and m (``Kb``),
    respectively.  Note ``Vcp`` is undefined for the zeroth embedding dimension (m = 0) 
    and due to the computational demand, **will take substantially more time to return function outputs.**
    See Appendix B in [2] for more info.
 
    .. code-block:: python
    
        Samp, A, B = SampEn(Sig, keyword = value, ...)
        
    Returns the sample entropy estimates (``Samp``) for dimensions = [0, 1, ..., ``m``]
    estimated for the data sequence (``Sig``) using the specified keyword arguments:
        :m:     - Embedding Dimension, a positive integer
        :tau:   - Time Delay, a positive integer
        :r:     - Radius Distance Threshold, a positive scalar  
        :Logx:  - Logarithm base, a positive scalar  
        :Vcp:   - Option to return variance of conditional probabilities and the number of overlapping matching vector pairs, a boolean [default: False]
 
    :See also:
        ``ApEn``, ``FuzzEn``, ``PermEn``, ``CondEn``, ``XSampEn``, ``SampEn2D``, ``MSEn``
  
    :References:
        [1] Joshua S Richman and J. Randall Moorman. 
            "Physiological time-series analysis using approximate entropy
            and sample entropy." 
            American Journal of Physiology-Heart and Circulatory Physiology 
            2000
            
        [2]  Douglas E Lake, Joshua S Richman, M.P. Griffin, J. Randall Moorman
            "Sample entropy analysis of neonatal heart rate variability."
            American Journal of Physiology-Regulatory, Integrative and Comparative Physiology
            283, no. 3 (2002): R789-R797.
    """
   
    Sig = np.squeeze(Sig)
    N = Sig.shape[0]  
    if r is None:
        r = 0.2*np.std(Sig)
  
    assert N>10 and Sig.ndim == 1,  "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m > 0), "m:     must be an integer > 0"
    assert isinstance(tau,int) and (tau > 0), "tau:   must be an integer > 0"
    assert isinstance(r,(int,float)) and (r>=0), "r:     must be a positive value"
    assert isinstance(Logx,(int,float)) and (Logx>0), "Logx:     must be a positive value"
    assert isinstance(Vcp,bool), "Vcp:     must be a Boolean"    

    Counter = (abs(np.expand_dims(Sig,axis=1)-np.expand_dims(Sig,axis=0))<= r)*np.triu(np.ones((N,N)),1)  
    M = np.hstack((m*np.ones(N-m*tau), np.repeat(np.arange(m-1,0,-1),tau)))
    A = np.zeros(m + 1)
    B = np.zeros(m + 1)
    A[0] = np.sum(Counter)
    B[0] = N*(N-1)/2
    
    for n in range(M.shape[0]):
        ix = np.where(Counter[n, :] == 1)[0]
        
        for k in range(1,int(M[n]+1)):              
            ix = ix[ix + (k*tau) < N]
            p1 = np.tile(Sig[n: n+1+(tau*k):tau], (ix.shape[0], 1))                       
            p2 = Sig[np.expand_dims(ix,axis=1) + np.arange(0,(k*tau)+1,tau)]
            ix = ix[np.amax(abs(p1 - p2), axis=1) <= r] 
            if ix.shape[0]:
                Counter[n, ix] += 1
            else:
                break
        
    for k in range(1, m+1):
        A[k] = np.sum(Counter > k)
        B[k] = np.sum(Counter[:,:-(k*tau)] >= k)
        
    with np.errstate(divide='ignore', invalid='ignore'):
        Samp = -np.log(A/B)/np.log(Logx) 
   
    
    if Vcp:
        Temp = np.vstack(np.where(Counter>m)).T
        if len(Temp)>1:
            Ka = np.zeros(len(Temp)-1)             
            for k in range(len(Temp)-1):
                TF = (abs(Temp[k+1:,:] - Temp[k,0]) <= m*tau) + (abs(Temp[k+1:,:] - Temp[k,1]) <= m*tau)
                Ka[k] = TF.any(axis=1).sum()
        else:
            Ka = 0
        
        Temp = np.vstack(np.where(Counter[:,:-m*tau] >=m)).T        
        if len(Temp)>1:
            Kb = np.zeros(len(Temp)-1)   
            for k in range(len(Temp)-1):
                TF = (abs(Temp[k+1:,:] - Temp[k,0]) <= (m-1)*tau) + (abs(Temp[k+1:,:] - Temp[k,1]) <= (m-1)*tau)
                Kb[k] = TF.any(axis=1).sum()      
        else:
            Kb = 0
            
        Ka = np.sum(Ka)
        Kb = np.sum(Kb)
        CP = A[-1]/B[-1]
        Vcp = (CP*(1-CP)/B[-1]) + (Ka - Kb*(CP**2))/(B[-1]**2)

        return Samp, A, B, (Vcp, Ka, Kb)
    
    else:
        return Samp, A, B        

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""


"""Base Kolmogorov Entropy function."""
import numpy as np

def K2En(Sig, m=2, tau=1, r=None, Logx=np.exp(1)):
    """K2En  estimates the Kolmogorov (K2) entropy of a univariate data sequence.

    .. code-block:: python
    
        K2, Ci = K2En(Sig) 
    
    Returns the Kolmogorov entropy estimates (``K2``) and the correlation
    integrals (``Ci``) for ``m`` = [1,2] estimated from the data sequence (``Sig``)
    using the default parameters: embedding dimension = 2, time delay = 1, 
    distance threshold (``r``) = 0.2*SD(``Sig``), logarithm = natural
 
    .. code-block:: python
     
        K2, Ci = K2En(Sig, keyword = value, ...)
        
    Returns the Kolmogorov entropy estimates (``K2``) for dimensions = [1, ..., ``m``]
    estimated from the data sequence (``Sig``) using the 'keyword' arguments:
        :m:    - Embedding Dimension, a positive integer
        :tau:   - Time Delay, a positive integer
        :r:     - Radius Distance Threshold, a positive scalar  
        :Logx:  - Logarithm base, a positive scalar  
 
    :See also: 
        ``DistEn``, ``XK2En``, ``MSEn``    
      
    :References:
        [1] Peter Grassberger and Itamar Procaccia,
            "Estimation of the Kolmogorov entropy from a chaotic signal." 
            Physical review A 28.4 (1983): 2591.
    
        [2] Lin Gao, Jue Wang  and Longwei Chen
            "Event-related desynchronization and synchronization 
            quantification in motor-related EEG by Kolmogorov entropy"
            J Neural Eng. 2013 Jun;10(3):03602
    
    """
    
    Sig = np.squeeze(Sig)
    N = Sig.shape[0]    
    if r is None:
        r = 0.2*np.std(Sig)
    
    assert N>10 and Sig.ndim == 1, "Sig:   must be a numpy vector"
    assert isinstance(m,int) and (m>0), "m:     must be an integer > 0"
    assert isinstance(tau,int) and (tau>0), "tau:   must be an integer > 0"
    assert isinstance(r,(int,float)) and r>0, "r:     must be a positive value"
    assert isinstance(Logx,(int,float)) and Logx>0, "Logx:     must be a positive value"
               
    m = m+1
    Zm = np.zeros((N,m))  
    Ci = np.zeros(m)
    for n in range(m):
        N2 = N - n*tau
        Zm[:N2,n] = Sig[n*tau:]  
        Norm = np.ones((N2-1,N2-1))*np.inf     
        for k in range(N2-1):
            Temp = np.tile(Zm[k,:n+1],(N2-k-1,1)) - Zm[k+1:N2,:n+1]
            Norm[k,k:] = np.linalg.norm(Temp,axis=1)   
        # Norm[Norm==0] = np.inf;
        Ci[n] = 2*np.sum(Norm < r)/(N2*(N2-1))
    
    with np.errstate(divide='ignore', invalid='ignore'):
        K2 = (np.log(Ci[:-1]/Ci[1:])/np.log(Logx))/tau
    K2[np.isinf(K2)] = np.NaN    
    return  K2, Ci

"""
    Copyright 2024 Matthew W. Flood, EntropyHub
    
    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at
    
     http://www.apache.org/licenses/LICENSE-2.0
    
    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
    
    For Terms of Use see https://github.com/MattWillFlood/EntropyHub
"""